from typing import (
    Dict,
    Optional,
)

from . import v1
from .api import API
from .resource import Resource
from .transport import (
    HTTPWireProtocol,
    WireProtocol,
)


# Mypy does not allow using modules as subtypes of protocols:
# https://github.com/python/mypy/issues/5018
supported_versions: Dict[str, API] = {"1": v1}  # type: ignore


class AlmaClient:
    """HTTP client for the Ex Libris Alma REST API."""

    api: API
    transport: WireProtocol

    _resource_cache: Dict[str, Resource] = {}

    def __init__(
        self, url: str, headers: Optional[Dict[str, str]] = None, version: str = "1"
    ):
        """
        Builds a new Alma API connection.

        The optional ``version`` parameter decides which version
        of the Zoom API to use.  Supported API versions are listed
        in ``dezoom.zoom.client.supported_versions``.

        :raises ValueError:
            If the provided ``url`` is not valid,
            or if the ``version`` is not supported.
        """
        try:
            self.api = supported_versions[version]
        except KeyError:
            raise ValueError(f"Unsupported API version: {version}")
        self.transport = HTTPWireProtocol(url, headers=headers)

    def __getattr__(self, name):
        try:
            return self._ensure_resource(name)
        except KeyError as err:
            raise AttributeError() from err

    def _ensure_resource(self, ns) -> Resource:
        if ns in self._resource_cache:
            return self._resource_cache[ns]
        cls = self.api.resources[ns]
        self._resource_cache[ns] = cls(transport=self.transport)
        return self._ensure_resource(ns)

    def __str__(self):
        return f"<{self.__class__.__name__} {self.transport.url}>"
