from collections import defaultdict
from typing import Dict, Optional, Protocol, Literal, Any, Union
from urllib.parse import (
    urljoin,
    urlparse,
)

import requests

Method = Literal["GET", "HEAD", "POST", "PUT", "DELETE", "PATCH"]


class WireProtocol(Protocol):
    def get(
        self,
        path: str,
        query: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        ...

    def post(
        self,
        path: str,
        data: Dict[str, Any],
        query: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        ...

    def put(
        self,
        path: str,
        data: Dict[str, Any],
        query: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        ...

    url: str


class HTTPWireProtocol(WireProtocol):
    def __init__(
        self,
        url: str,
        prefix: str = "",
        headers: Union[Dict[str, str], None] = None,
    ):
        self.url = urljoin(url, prefix)
        self.headers = defaultdict(str, headers or {})
        self.session = requests.Session()

    def __str__(self):
        return f"<{type(self).__name__} {self.url} object at {hex(id(self))}>"

    def __repr__(self):
        return (
            f"{self.__module__}.{type(self).__name__}("
            f"{self.url!r}, "
            f"headers={dict(self.headers)!r}"
            ")"
        )

    def url_for(self, path: str) -> str:
        return urljoin(self.url, path)

    def get(
        self,
        path: str,
        query: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        return self.request("GET", path, query)

    def post(
        self,
        path: str,
        data: Dict[str, Any],
        query: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        return self.request("POST", path, query, json=data)

    def put(
        self,
        path: str,
        data: Dict[str, Any],
        query: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        return self.request("PUT", path, query, json=data)

    def request(
        self,
        method: Method,
        path: str,
        params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], bytes]] = None,
    ) -> requests.Response:
        resp = self._raw_request(method, path, params=params, json=json)
        # resp.raise_for_status()
        return resp

    def _raw_request(
        self,
        method: Method,
        path: str,
        params: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], bytes]] = None,
    ) -> requests.Response:
        kwargs = {}
        if isinstance(json, bytes):
            kwargs["data"] = json
        else:
            kwargs["json"] = json
        return self.session.request(
            method,
            self.url_for(path),
            params=params,
            headers=self.headers,
            **kwargs,
        )
