"""
Methods for accessing the code tables API.

https://developers.exlibrisgroup.com/blog/Working-with-the-code-tables-API/
"""

import logging
from typing import List, Optional

from pydantic import BaseModel

from ..resource import Resource
from ..utils import quote_path_arg
from .common import DescribedValue

logger = logging.getLogger(__name__)


class CodeTableScope(BaseModel):
    institution_id: DescribedValue
    library_id: DescribedValue


class Code(BaseModel):
    code: str
    description: str
    default: bool
    enabled: Optional[bool]


class CodeTableBase(BaseModel):
    name: str
    description: str
    sub_system: DescribedValue
    patron_facing: bool
    language: Optional[DescribedValue]


class CodeTable(CodeTableBase):
    scope: CodeTableScope
    row: List[Code]


class CodeTableListItem(BaseModel):
    link: str


class CodeTableList(BaseModel):
    code_table: List[CodeTableListItem]
    total_record_count: int


class CodeTables(Resource):
    def get(self, name: str) -> Optional[CodeTable]:
        """
        Fetch a single code table by name.

        Known code table names:
          - UserGroups
          - Genders
        """
        url = "conf/code-tables/{name}".format(name=quote_path_arg(name))
        response = self.transport.get(url)
        data = response.json()
        return CodeTable.parse_obj(data)

    def list(self) -> Optional[CodeTableList]:
        """
        List all code tables.
        """
        url = "conf/code-tables"
        response = self.transport.get(url)
        data = response.json()
        return CodeTableList.parse_obj(data)
