import requests
from typing import Type, Dict, Union, Optional, TypeVar

from .errors import (
    AlmaTooManyResults,
    AlmaNotFound,
)
from .common import SearchWrapper
from .utils import response2object
from ..resource import Resource

TAlmaModel = TypeVar("TAlmaModel", bound="AlmaModel")


class AlmaResource(Resource):
    """Represents an API resource."""

    def _list_page(
        self,
        url: str,
        limit: int = 20,
        offset: int = 0,
        q: Optional[str] = None,
    ) -> requests.Response:
        """List a page of results"""
        params: Dict[str, Union[str, int]] = {"limit": limit, "offset": offset}
        if q:
            params["q"] = q
        return self.transport.get(url, query=params)

    def _find_resource(
        self, search_wrapper: Type[SearchWrapper], url: str, q: str
    ) -> TAlmaModel:
        """
        Find resource
        """
        response = self._list_page(url=url, q=q, limit=1)
        page: SearchWrapper = response2object(response, search_wrapper)

        if page.total_record_count == 1:
            return page.get_results()[0]
        if page.total_record_count > 1:
            raise AlmaTooManyResults(f"Too many results: {page.total_record_count}")
        raise AlmaNotFound("No results found")

    def _list_resource(
        self,
        search_wrapper: Type[SearchWrapper],
        url: str,
        limit: int = 20,
        offset: int = 0,
        q: Optional[str] = None,
    ):
        """
        List resources

        :param limit: Limits the number of results
        :param offset: Offset of the results returned.
        :param q:
            Search query. Optional.
            Note that some search query terms are different from
            the corresponding field name. Eg field: searchable_id -> searchable_ids
        """
        params = {
            "url": url,
            "limit": limit,
            "offset": offset,
            "q": q,
        }

        response = self._list_page(**params)
        page: SearchWrapper = response2object(response, search_wrapper)
        results = page.get_results()
        if results is None:
            return iter([])
        while results:
            for obj in results:
                yield obj
            if (limit + offset) >= page.total_record_count:
                break
            params["offset"] += limit
            response = self._list_page(**params)
            page: SearchWrapper = response2object(response, search_wrapper)
            results = page.get_results()
