import logging
from typing import Optional, List, Dict, Any

import requests
from pydantic import BaseModel

from .common import DescribedValue, SearchWrapper
from .resource import AlmaResource
from .utils import get_single_object
from ..utils import quote_path_arg

logger = logging.getLogger(__name__)


class OptionallyDescribedValue(BaseModel):
    value: str
    desc: Optional[str]


class EmailContact(BaseModel):
    email_address: str
    description: Optional[str]
    preferred: bool
    segment_type: str  # e.g. "External"
    email_type: List[DescribedValue]


class PhoneContact(BaseModel):
    phone_number: str
    preferred: bool
    preferred_sms: bool
    segment_type: str  # e.g. "External"
    phone_type: List[DescribedValue]


class AddressContact(BaseModel):
    line1: str
    line2: Optional[str]
    line3: Optional[str]
    line4: Optional[str]
    line5: Optional[str]
    city: Optional[str]
    country: DescribedValue
    state_province: Optional[str]
    postal_code: Optional[str]
    address_note: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]
    address_type: List[DescribedValue]
    preferred: bool
    segment_type: str


class ContactInfo(BaseModel):
    address: List[AddressContact]
    email: List[EmailContact]
    phone: List[PhoneContact]


class UserIdentifier(BaseModel):
    value: str
    id_type: DescribedValue
    note: Optional[str]
    status: str  # e.g. "ACTIVE"
    segment_type: str  # e.g. "External"


class UserRoleParameter(BaseModel):
    type: OptionallyDescribedValue
    scope: Optional[DescribedValue]
    value: OptionallyDescribedValue


class UserRole(BaseModel):
    status: DescribedValue
    scope: DescribedValue
    role_type: DescribedValue
    parameter: List[UserRoleParameter]


class UserBlock(BaseModel):
    block_type: DescribedValue
    block_description: DescribedValue
    block_status: str
    block_note: str
    created_by: str
    created_date: str  # 1990-01-30Z
    expiry_date: str  # 1990-01-30Z
    item_loan_id: str
    block_owner: str


class UserNote(BaseModel):
    note_type: DescribedValue
    note_text: str
    user_viewable: bool
    popup_note: bool
    created_by: str
    created_date: str  # 1990-01-30Z
    note_owner: Optional[str]


class UserStatistic(BaseModel):
    statistic_category: DescribedValue
    category_type: DescribedValue
    statistic_owner: str
    statistic_note: str


class UserProxy(BaseModel):
    primary_id: str
    full_name: str
    proxy_type: str


class LibraryNotice(BaseModel):
    code: DescribedValue
    value: DescribedValue


class RsLibrary(BaseModel):
    code: DescribedValue


class User(BaseModel):
    primary_id: str
    first_name: str
    middle_name: str
    last_name: str
    full_name: str
    record_type: DescribedValue
    user_title: DescribedValue
    job_category: DescribedValue
    job_description: str
    gender: DescribedValue
    user_group: DescribedValue
    campus_code: DescribedValue
    web_site_url: str
    cataloger_level: DescribedValue
    preferred_language: DescribedValue
    birth_date: Optional[str]  # 1990-01-30Z
    expiry_date: Optional[str]  # 1990-01-30Z
    account_type: DescribedValue
    external_id: str  # e.g. "BULK_UPDATE"
    password: str
    force_password_change: str
    status: DescribedValue
    status_date: str  # 2020-01-30Z
    last_patron_activity_date: Optional[str]  # 2020-01-30Z
    # requests: Optional[?]
    # loans: Optional[?]
    # fees: Optional[?]
    contact_info: ContactInfo
    linking_id: Optional[str]
    pref_first_name: str
    pref_middle_name: str
    pref_last_name: str
    pref_name_suffix: str
    is_researcher: Optional[bool]
    # researcher: Optional[?]
    # link: Optional[?]
    user_identifier: List[UserIdentifier]
    user_role: List[UserRole]
    user_block: List[UserBlock]
    user_note: List[UserNote]
    user_statistic: List[Any]
    # proxy_for_user: List[UserProxy]
    rs_library: List[RsLibrary]
    # library_notice: List[LibraryNotice]


class UserSearchResult(BaseModel):
    primary_id: str
    first_name: str
    last_name: str
    gender: DescribedValue
    password: str
    status: DescribedValue
    # requests: Optional[?]
    # loans: Optional[?]
    # fees: Optional[?]
    is_researcher: Optional[bool]
    # researcher: Optional[?]
    # link: Optional[?]


class UserSearchWrapper(SearchWrapper):
    user: Optional[List[UserSearchResult]]

    def get_results(self) -> Optional[List[UserSearchResult]]:
        return self.user


class Users(AlmaResource):
    @get_single_object(User)
    def get(
        self, user_id: str, user_id_type: Optional[str] = None
    ) -> requests.Response:
        """
        Fetch a single user.

        Relevant user identifier types:
          - all_unique (default on the server side if nothing is specified, matches all of the below)
          - (todo: this one is a lie) PRIMARYIDENTIFIER (equals primary_id on user objects)
          - UNIV_ID ("University ID", should be uid@example.com)
          - BARCODE (Physical card barcode)
        """
        url = "users/{user_id}".format(user_id=quote_path_arg(user_id))
        params = {}
        if user_id_type:
            params["user_id_type"] = user_id_type
        return self.transport.get(url, query=params)

    @get_single_object(User)
    def create(self, data: Dict[str, Any]) -> requests.Response:
        """
        Create a new user.
        """
        url = "users"
        return self.transport.post(url, data=data)

    @get_single_object(User)
    def update(
        self,
        user_id: str,
        data: Dict[str, Any] = None,
        user_id_type: Optional[str] = None,
        override: Optional[List[str]] = None,
    ) -> requests.Response:
        """
        Update a single user.

        :param user_id_type:
            Defaults to "all_unique", otherwise refer to the user identifier type
            code table.

        :param override:
          The following fields are not updated unless specified in this override
          parameter:
             user_group, job_category, pin_number, preferred_language, campus_code,
             rs_libraries, user_title, library_notices,
             pref_first_name, pref_middle_name, pref_last_name, cataloger_level
        """
        url = "users/{user_id}".format(user_id=quote_path_arg(user_id))
        params = {"user_id_type": user_id_type}
        if override:
            assert isinstance(override, list)
            assert all([isinstance(x, str) for x in override])
            params["override"] = ",".join(override)
        return self.transport.put(url, query=params, data=data)

    def list(self, limit: int = 20, offset: int = 0, q: Optional[str] = None):
        """
        List users.

        :param limit: Limits the number of results
        :param offset: Offset of the results returned.
        :param q:
            Search query. Optional.
            Searching for words from:
              - primary_id
              - first_name
              - last_name
              - middle_name
              - email
              - job_category
              - identifiers
              - general_info (keywords from all fields except emails and identifiers) and
              - ALL (words from all fields but without support for the AND operator)

              Example (note the tilde between the code and text): q=last_name~Smith
        """
        return self._list_resource(
            UserSearchWrapper, "users", limit=limit, offset=offset, q=q
        )
