import requests
from typing import Type, Callable, TypeVar
from functools import wraps
from requests.exceptions import HTTPError

from .errors import handle_error

TBaseModel = TypeVar("TBaseModel", bound="BaseModel")


def response2object(
    response: requests.models.Response, model: Type[TBaseModel]
) -> TBaseModel:
    """Convert a request response into a populated pydantic object"""
    try:
        response.raise_for_status()
    except HTTPError as err:
        if response.status_code == 400:
            handle_error(response, err)
        raise
    data = response.json()
    return model.parse_obj(data)


def get_single_object(model: Type[TBaseModel]) -> Callable:
    """Convert the returned response of the decorated function into a pydantic object"""

    def decorate(func: Callable) -> Callable:
        @wraps(func)
        def wrapped(*args, **kwargs) -> TBaseModel:
            response = func(*args, **kwargs)
            return response2object(response, model)

        return wrapped

    return decorate
